var types = (function() { // Start types.js
    "use strict";

    var typesList = {
        "number": Number,
        "string": String,
        "Array": Array,
        "Object": Object
    };

    var setupTypeModern = function(object, name) {
        Object.defineProperty(object.prototype, "__type__", {
            value: name
        });
    };

    var setupTypeLegacy = function(object, name) {
        object.prototype.__type__ = name;
    };

    var types = function(method) {
        var params, returns;

        var typedMethod = function() {
            var res;

            if (types.enabled && params) {
                types.verify(params, arguments, 'parameter');
            }

            res = method.apply(this, arguments);

            if (types.enabled && returns) {
                types.verify(returns, res, 'return');
            }

            return res;
        };

        typedMethod.params = typedMethod.p = function() {
            params = arguments;

            return this;
        };

        typedMethod.returns = typedMethod.r = function() {
            returns = arguments;

            return this;
        };

        return typedMethod;
    };

    var createError = function(err) {
        if (typeof console === "object" && console.error) {
            console.error("Stacktrace @types");
        }

        throw new TypeError(err);
    };

    types.verify = function(types, values, context) {
        var type, value;

        for (var i in values) {
            if (!values.hasOwnProperty(i)) {
                continue;
            }

            type = types[i];
            value = values.length > 0 && values.__type__ !== "string" ? values[i] : values;

            if (typeof type === 'undefined') {
                return;
            }

            if (!((value instanceof type) ||
                  (value.__type__ === type.prototype.__type__))) {
                var message =
                    "in " + context + "." +

                    "\n\nExpected -> " + type.toString().match(/function (.*) ?\(/)[1] +
                    "\nActual -> " + value.constructor.toString().match(/function (.*) ?\(/)[1] +

                    "\n\n@types";

                createError(message);
            }
        }
    };

    types.setupTypes = function(types) {
        var name, object;

        for (var i in types) {
            if (!types.hasOwnProperty(i)) {
                continue;
            }

            object = types[i];
            name = i;

            if (typeof Object.defineProperty === "function") {
                setupTypeModern(object, name);
            } else {
                setupTypeLegacy(object, name);
            }
        }
    };

    types.setupTypes(typesList);

    types.enabled = true;

    return types;
})(); // End types.js